"""
Core semantic version functionality.

Author: Tom Fleet
Created: 08/10/2021
"""

from __future__ import annotations

import json
import re
import sys
from typing import Optional, Tuple

# Compatability with python 3.7
if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

# See https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
# The only thing we've added is the optional v at the start
_SEMVER_REGEX = re.compile(
    r"""^v?(?P<major>0|[1-9]\d*)\. # Major
    (?P<minor>0|[1-9]\d*)\. # Minor
    (?P<patch>0|[1-9]\d*) # Patch
    (?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))? # Optional pre-release # noqa: E501
    (?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$""",  # Optional build metadata
    flags=re.VERBOSE,
)


class VersionDict(TypedDict):
    """
    Schema for the dictionary a `Version` object
    expects to unpack
    """

    major: int
    minor: int
    patch: int
    prerelease: Optional[str]
    buildmetadata: Optional[str]


VersionTuple = Tuple[int, int, int, Optional[str], Optional[str]]


class Version:
    def __init__(
        self,
        major: int,
        minor: int,
        patch: int,
        prerelease: Optional[str] = None,
        buildmetadata: Optional[str] = None,
    ) -> None:
        """
        A data container for a semantic version.

        Args:
            major (int): The major version, to be incremented when making
                backwards incompatible (breaking) changes.
            minor (int): The minor version, to be incremented when introducing
                backwards compatible features.
            patch (int): The patch version, to be incremented when making
                backwards compatible bug fixes.
            prerelease (Optional[str], optional): Any pre-release tags
                e.g. 'rc.1'. Defaults to None.
            buildmetadata (Optional[str], optional): Any build meta tags
                e.g. 'build.123'. Defaults to None.

        Raises:
            ValueError: If any numeric version part < 0.
        """
        self.major = major
        self.minor = minor
        self.patch = patch
        self.prerelease = prerelease
        self.buildmetadata = buildmetadata

        if any(part < 0 for part in (self.major, self.minor, self.patch)):
            raise ValueError(
                f"Version {self!r} is invalid. Parts cannot be less than 0."
            )

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(major={self.major!r}, minor={self.minor!r}, patch={self.patch!r}, "
            + f"prerelease={self.prerelease!r}, buildmetadata={self.buildmetadata!r})"
        )

    def __str__(self) -> str:
        ver = f"v{self.major}.{self.minor}.{self.patch}"

        if self.prerelease:
            ver += f"-{self.prerelease}"

        if self.buildmetadata:
            ver += f"+{self.buildmetadata}"

        return ver

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Version):
            raise TypeError(f"Cannot compare object of type Version and {type(other)}")

        return (
            self.major,
            self.minor,
            self.patch,
            self.prerelease,
            self.buildmetadata,
        ) == (
            other.major,
            other.minor,
            other.patch,
            other.prerelease,
            other.buildmetadata,
        )

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Version):
            raise TypeError(f"Cannot compare object of type Version and {type(other)}")

        if (self.major, self.minor, self.patch) < (
            other.major,
            other.minor,
            other.patch,
        ):
            return True

        pre_release_comp = self._compare_prerelease(other)
        if pre_release_comp == -1:
            return True

        build_comp = self._compare_build(other)
        if build_comp == -1:
            return True

        return False

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Version):
            raise TypeError(f"Cannot compare object of type Version and {type(other)}")

        if (self.major, self.minor, self.patch) > (
            other.major,
            other.minor,
            other.patch,
        ):
            return True

        pre_release_comp = self._compare_prerelease(other)
        if pre_release_comp == 1:
            return True

        build_comp = self._compare_build(other)
        if build_comp == 1:
            return True

        return False

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Version):
            raise TypeError(f"Cannot compare object of type Version and {type(other)}")

        if self == other:
            return True

        if self < other:
            return True

        return False

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Version):
            raise TypeError(f"Cannot compare object of type Version and {type(other)}")

        if self == other:
            return True

        if self > other:
            return True

        return False

    def _compare_prerelease(self, other: Version) -> int:
        """
        Helper to try and compare the pre-release versions.

        It will check for string equality, whether or not
        the different versions have pre-releases, check
        for integers in the text and compare based on them,
        and if none of this works will raise a ValueError.

        Note: This is only called if the numeric version
        is equal.
        """
        if self.prerelease == other.prerelease:
            # The pre release strings are equal
            return 0
        elif self.prerelease and not other.prerelease:
            # Ours has a pre-release but the other doesn't
            # meaning ours is less
            return -1
        elif not self.prerelease and other.prerelease:
            # Ours doesn't have a pre-release but the other
            # does meaning ours is greater
            return 1

        # Try and extract a number to compare
        if self.prerelease and other.prerelease:
            ours = re.findall(pattern=r"\d+", string=self.prerelease)
            others = re.findall(pattern=r"\d+", string=other.prerelease)

            if not ours or not others:
                raise ValueError("Could not parse comparable pre-release version info.")

            if len(ours) > 1 or len(others) > 1:
                raise ValueError("Multiple integers found in pre-release version.")

            our_version = int(ours.pop())
            other_version = int(others.pop())

            if our_version == other_version:
                return 0
            elif our_version > other_version:
                return 1
            elif our_version < other_version:
                return -1

        # If we get here, we couldn't parse the pre-release
        raise ValueError(f"Could not compare {self.prerelease} and {other.prerelease}")

    def _compare_build(self, other: Version) -> int:
        """
        Helper to try and compare the build metadata versions.

        It will check for string equality, whether or not
        the different versions have pre-releases, check
        for integers in the text and compare based on them,
        and if none of this works will raise a ValueError.

        Note: This is only called if both the numeric version
        and the pre-release strings are equal.
        """
        if self.buildmetadata == other.buildmetadata:
            # The pre release strings are equal
            return 0
        elif self.buildmetadata and not other.buildmetadata:
            # Ours has a pre-release but the other doesn't
            # meaning ours is less
            return -1
        elif not self.buildmetadata and other.buildmetadata:
            # Ours doesn't have a pre-release but the other
            # does meaning ours is greater
            return 1

        # Try and extract a number to compare
        if self.buildmetadata and other.buildmetadata:
            ours = re.findall(pattern=r"\d+", string=self.buildmetadata)
            others = re.findall(pattern=r"\d+", string=other.buildmetadata)

            if not ours or not others:
                raise ValueError(
                    "Could not parse comparable build metadata version info."
                )

            if len(ours) > 1 or len(others) > 1:
                raise ValueError("Multiple integers found in build metadata version.")

            our_version = int(ours.pop())
            other_version = int(others.pop())

            if our_version == other_version:
                return 0
            elif our_version > other_version:
                return 1
            elif our_version < other_version:
                return -1

        # If we get here, we couldn't parse the pre-release
        raise ValueError(
            f"Could not compare {self.buildmetadata} and {other.buildmetadata}"
        )

    def is_valid(self) -> bool:
        """
        Checks the `Version` against the official
        semver regex pattern and reports whether or not it
        is a valid semver.

        Returns:
            bool: True if `Version` is valid, else False

        Examples:

        ```python
        >>> v = Version(1, 2, 4)
        >>> v.is_valid()
        True
        ```

        ```python
        >>> v = Version(1, 2, 4, "blah198y_+-2-", "build---19790")
        >>> v.is_valid()
        False
        ```
        """
        return bool(_SEMVER_REGEX.match(self.to_string()))

    def bump_major(self) -> Version:
        """
        Return a new `Version` with the major version
        number bumped.

        Returns:
            Version: New bumped version.

        Examples:

        ```python
        >>> v1 = Version(1, 2, 4)
        >>> v1.bump_major()
        Version(major=2, minor=0, patch=0, prerelease=None, buildmetadata=None)
        ```

        ```python
        >>> v1 = Version(0, 7, 6, "rc.1", "build.123")
        >>> v1.bump_major()
        Version(major=1, minor=0, patch=0, prerelease=None, buildmetadata=None)
        ```
        """
        return Version(self.major + 1, 0, 0)

    def bump_minor(self) -> Version:
        """
        Return a new `Version` with the minor version
        number bumped.

        Returns:
            Version: New bumped version.

        Examples:

        ```python
        >>> v1 = Version(1, 2, 4)
        >>> v1.bump_minor()
        Version(major=1, minor=3, patch=0, prerelease=None, buildmetadata=None)
        ```

        ```python
        >>> v1 = Version(0, 7, 6, "rc.1", "build.123")
        >>> v1.bump_major()
        Version(major=0, minor=8, patch=0, prerelease=None, buildmetadata=None)
        ```
        """
        return Version(self.major, self.minor + 1, 0)

    def bump_patch(self) -> Version:
        """
        Return a new `Version` with the patch version
        number bumped.

        Returns:
            Version: New bumped version.

        Examples:

        ```python
        >>> v1 = Version(1, 2, 4)
        >>> v1.bump_patch()
        Version(major=1, minor=2, patch=5, prerelease=None, buildmetadata=None)
        ```

        ```python
        >>> v1 = Version(0, 7, 6, "rc.1", "build.123")
        >>> v1.bump_major()
        Version(major=0, minor=7, patch=7, prerelease=None, buildmetadata=None)
        ```
        """
        return Version(self.major, self.minor, self.patch + 1)

    def to_string(self) -> str:
        """
        Generate a string representation of the
        `Version`.

        Returns:
            str: Version string.

        Examples:

        ```python
        >>> v = Version(1, 2, 4)
        >>> v.to_string()
        'v1.2.4'
        ```

        ```python
        >>> v = Version(1, 2, 4, "rc.2", "build.6")
        >>> v.to_string()
        'v1.2.4-rc.2+build.6'
        ```
        """
        return str(self)

    def to_tuple(self) -> VersionTuple:
        """
        Return the `Version` as a tuple of it's
        fields.

        Returns:
            VersionTuple: The Version tuple.

        Examples:

        ```python
        >>> v = Version(1, 2, 4)
        >>> v.to_tuple()
        (1, 2, 4, None, None)
        ```

        ```python
        >>> v = Version(1, 2, 4, "rc.2", "build.6")
        >>> v.to_tuple()
        (1, 2, 4, "rc.2", "build.6")
        ```
        """
        return (self.major, self.minor, self.patch, self.prerelease, self.buildmetadata)

    def to_dict(self) -> VersionDict:
        """
        Return the `Version` as a dictionary.

        Returns:
            VersionDict: The Version dictionary.

        Examples:

        ```python
        >>> v = Version(1, 2, 4)
        >>> v.to_dict()
        {'major': 1, 'minor': 2, 'patch': 4, 'prerelease': None, 'buildmetadata': None}
        ```

        ```python
        >>> v = Version(1, 2, 4, "rc.1", "build.2")
        >>> v.to_dict()
        {'major': 1, 'minor': 2, 'patch': 4, 'prerelease': 'rc.1', 'buildmetadata': 'build.2'}
        ```
        """
        return {
            "major": self.major,
            "minor": self.minor,
            "patch": self.patch,
            "prerelease": self.prerelease,
            "buildmetadata": self.buildmetadata,
        }

    def to_json(self) -> str:
        """
        Return the `Version` as a JSON string.

        Returns:
            str: The Version JSON string.

        Examples:

        ```python
        >>> v = Version(1, 2, 4)
        >>> v.to_json()
        '{"major": 1, "minor": 2, "patch": 4, "prerelease": null, "buildmetadata": null}'
        ```

        ```python
        >>> v = Version(1, 2, 4, "rc.1", "build.2")
        >>> v.to_json()
        '{"major": 1, "minor": 2, "patch": 4, "prerelease": "rc.1", "buildmetadata": "build.2"}'
        ```
        """
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, version_dict: VersionDict) -> Version:
        """
        Construct and return a `Version` from a dictionary
        of it's parts.

        Expects a dictionary with keys: `major`, `minor`, `patch`,
        `prerelease` and `buildmetadata`.

        Args:
            version_dict (VersionDict): Version as a dictionary of it's
                parts.

        Returns:
            Version: Constructed Version.

        Raises:
            TypeError: If the passed dictionary does not
                have keys matching the required parts.

        Examples:

        ```python
        >>> v = {"major": 1, "minor": 2, "patch": 4}
        >>> Version.from_dict(v)
        Version(major=1, minor=2, patch=4, prerelease=None, buildmetadata=None)
        ```

        ```python
        >>> v = {"major": 1, "minor": 2, "patch": 4, "prerelease": "rc.1", "buildmetadata": "build.123"}
        >>> Version.from_dict(v)
        Version(major=1, minor=2, patch=4, prerelease='rc.1', buildmetadata='build.123')
        ```
        """
        return Version(**version_dict)

    @classmethod
    def from_string(cls, string: str) -> Version:
        """
        Construct and return a `Version` from a valid semver
        string.

        Args:
            string (str): The semver string.

        Raises:
            ValueError: If the semver string does not
                pass the official semver regex.

        Returns:
            Version: The constructed Version.

        Examples:

        ```python
        >>> Version.from_string("v1.2.4")
        Version(major=1, minor=2, patch=4, prerelease=None, buildmetadata=None)
        ```

        ```python
        >>> Version.from_string("v1.2.4-rc.1+build.123")
        Version(major=1, minor=2, patch=4, prerelease='rc.1', buildmetadata='build.123')
        ```
        """
        match = _SEMVER_REGEX.match(string)
        if not match:
            raise ValueError(f"{string!r} is not a valid semver string.")

        return Version.from_dict(
            VersionDict(
                major=int(match.group("major")),
                minor=int(match.group("minor")),
                patch=int(match.group("patch")),
                prerelease=match.group("prerelease"),
                buildmetadata=match.group("buildmetadata"),
            )
        )

    @classmethod
    def from_tuple(cls, tup: VersionTuple) -> Version:
        """
        Construct and return a `Version` from a tuple of it's
        parts.

        Args:
            tup (VersionTuple): The tuple to construct the Version
                from.

        Returns:
            Version: The constructed Version.

        Examples:

        ```python
        >>> v = (1, 2, 4)
        >>> Version.from_tuple(v)
        Version(major=1, minor=2, patch=4, prerelease=None, buildmetadata=None)
        ```

        ```python
        >>> v = (1, 2, 4, "rc.1", "build.123")
        >>> Version.from_tuple(v)
        Version(major=1, minor=2, patch=4, prerelease='rc.1', buildmetadata='build.123')
        ```
        """
        return Version(*tup)

    @classmethod
    def from_json(cls, json_string: str) -> Version:
        """
        Construct and return a `Version` from a json string
        of it's parts.

        Args:
            json_string (str): The json string.

        Returns:
            Version: The constructed Version.

        Examples:

        ```python
        >>> v = '{"major": 1, "minor": 2, "patch": 4}'
        >>> Version.from_json(v)
        Version(major=1, minor=2, patch=4, prerelease=None, buildmetadata=None)
        ```

        ```python
        >>> v = '{"major": 1, "minor": 2, "patch": 4, "prerelease": "rc.1", "buildmetadata": "build.123"}'
        >>> Version.from_json(v)
        Version(major=1, minor=2, patch=4, prerelease='rc.1', buildmetadata='build.123')
        ```
        """
        data: VersionDict = json.loads(json_string)
        return Version(**data)



"""
Tests for the core functionality.

Author: Tom Fleet
Created: 09/10/2021
"""


import pytest

from madonna import Version
from madonna.version import VersionDict, VersionTuple


def test_version_init():
    v = Version(1, 2, 4, "rc.1", "build.123")
    assert v.major == 1
    assert v.minor == 2
    assert v.patch == 4
    assert v.prerelease == "rc.1"
    assert v.buildmetadata == "build.123"


def test_version_repr():
    v = Version(1, 2, 4, "rc.1", "build.123")
    want = (
        "Version(major=1, minor=2, patch=4, prerelease='rc.1',"
        " buildmetadata='build.123')"
    )
    assert repr(v) == want


def test_version_less_than_zero_major():
    """
    Initialising a version with any part < 0 should raise
    a ValueError.
    """
    with pytest.raises(ValueError):
        Version(-1, 2, 4)


def test_version_less_than_zero_minor():
    """
    Initialising a version with any part < 0 should raise
    a ValueError.
    """
    with pytest.raises(ValueError):
        Version(1, -2, 4)


def test_version_less_than_zero_patch():
    """
    Initialising a version with any part < 0 should raise
    a ValueError.
    """
    with pytest.raises(ValueError):
        Version(1, 2, -4)


@pytest.mark.parametrize(
    "version, want",
    [
        (Version(1, 2, 4, None, None), "v1.2.4"),
        (Version(1, 2, 4), "v1.2.4"),
        (Version(2, 6, 8, "rc.2", None), "v2.6.8-rc.2"),
        (Version(2, 6, 8, "rc.2"), "v2.6.8-rc.2"),
        (Version(7, 6, 2, None, "build.123"), "v7.6.2+build.123"),
        (Version(7, 2, 1, "rc.1", "build.123"), "v7.2.1-rc.1+build.123"),
    ],
)
def test_version_str(version: Version, want: str):
    assert str(version) == want


@pytest.mark.parametrize(
    "v1, v2, want",
    [
        (Version(1, 2, 3), Version(1, 2, 3), True),
        (Version(1, 2, 3, "pre"), Version(1, 2, 3, "pre"), True),
        (Version(1, 2, 3, "pre", "build"), Version(1, 2, 3, "pre", "build"), True),
        (Version(1, 2, 3, None, None), Version(1, 2, 3, None, None), True),
        (Version(0, 2, 3), Version(1, 2, 3), False),
        (Version(1, 1, 3), Version(1, 2, 3), False),
        (Version(1, 2, 2), Version(1, 2, 3), False),
        (Version(1, 2, 3, "pre"), Version(1, 2, 3), False),
        (Version(1, 2, 3, None, "build"), Version(1, 2, 3, None, None), False),
        (
            Version(1, 2, 3, "pre", "build"),
            Version(1, 2, 3, "diffpre", "diffbuild"),
            False,
        ),
    ],
)
def test_version_eq(v1: Version, v2: Version, want: bool):
    assert (v1 == v2) is want


def test_version_eq_invalid_type():
    v = Version(1, 2, 4)

    with pytest.raises(TypeError):
        v == "a string"


@pytest.mark.parametrize(
    "v1, v2, want",
    [
        (Version(1, 2, 3), Version(1, 2, 3), False),
        (Version(1, 2, 3, "pre"), Version(1, 2, 3, "pre"), False),
        (Version(1, 2, 3, "pre", "build"), Version(1, 2, 3, "pre", "build"), False),
        (Version(1, 2, 3, None, None), Version(1, 2, 3, None, None), False),
        (Version(0, 2, 3), Version(1, 2, 3), True),
        (Version(1, 1, 3), Version(1, 2, 3), True),
        (Version(1, 2, 2), Version(1, 2, 3), True),
        (Version(1, 2, 3, "pre"), Version(1, 2, 3), True),
        (Version(1, 2, 3, None, "build"), Version(1, 2, 3, None, None), True),
        (
            Version(1, 2, 3, "pre", "build"),
            Version(1, 2, 3, "diffpre", "diffbuild"),
            True,
        ),
    ],
)
def test_version_ne(v1: Version, v2: Version, want: bool):
    assert (v1 != v2) is want


@pytest.mark.parametrize(
    "v1, v2, want",
    [
        (Version(0, 7, 6), Version(1, 7, 6), True),
        (Version(1, 6, 6), Version(1, 7, 6), True),
        (Version(1, 7, 5), Version(1, 7, 6), True),
        (Version(1, 7, 6, "pre"), Version(1, 7, 6), True),
        (Version(1, 7, 6, "pre", "build"), Version(1, 7, 6), True),
        (Version(1, 7, 6, "pre1"), Version(1, 7, 6, "pre2"), True),
        (Version(1, 7, 6, "pre1", "build1"), Version(1, 7, 6, "pre1", "build2"), True),
        (Version(1, 7, 6, None, "build1"), Version(1, 7, 6, None, "build2"), True),
        (Version(1, 7, 6), Version(0, 7, 6), False),
        (Version(1, 7, 6), Version(1, 6, 6), False),
        (Version(1, 7, 6), Version(1, 7, 5), False),
        (Version(1, 7, 6), Version(1, 7, 6, "pre"), False),
        (Version(1, 7, 6), Version(1, 7, 6, "pre", "build"), False),
        (Version(1, 7, 6, "pre2"), Version(1, 7, 6, "pre1"), False),
        (Version(1, 7, 6, "pre2", "build2"), Version(1, 7, 6, "pre1", "build1"), False),
        (Version(1, 7, 6, None, "build2"), Version(1, 7, 6, None, "build1"), False),
    ],
)
def test_version_lt(v1: Version, v2: Version, want: bool):
    assert (v1 < v2) is want


def test_version_lt_invalid_type():
    v = Version(1, 2, 4)

    with pytest.raises(TypeError):
        v < "a string"


@pytest.mark.parametrize(
    "v1, v2, want",
    [
        (Version(0, 7, 6), Version(1, 7, 6), True),
        (Version(1, 6, 6), Version(1, 7, 6), True),
        (Version(1, 7, 5), Version(1, 7, 6), True),
        (Version(1, 7, 6), Version(1, 7, 6), True),
        (Version(1, 7, 6, "pre"), Version(1, 7, 6), True),
        (Version(1, 7, 6, "pre", "build"), Version(1, 7, 6), True),
        (Version(1, 7, 6, "pre1"), Version(1, 7, 6, "pre2"), True),
        (Version(1, 7, 6, "pre1", "build1"), Version(1, 7, 6, "pre1", "build2"), True),
        (Version(1, 7, 6, None, "build1"), Version(1, 7, 6, None, "build2"), True),
        (Version(1, 7, 6), Version(0, 7, 6), False),
        (Version(1, 7, 6), Version(1, 6, 6), False),
        (Version(1, 7, 6), Version(1, 7, 5), False),
        (Version(1, 7, 6), Version(1, 7, 6, "pre"), False),
        (Version(1, 7, 6), Version(1, 7, 6, "pre", "build"), False),
        (Version(1, 7, 6, "pre2"), Version(1, 7, 6, "pre1"), False),
        (Version(1, 7, 6, "pre2", "build2"), Version(1, 7, 6, "pre1", "build1"), False),
        (Version(1, 7, 6, None, "build2"), Version(1, 7, 6, None, "build1"), False),
    ],
)
def test_version_le(v1: Version, v2: Version, want: bool):
    assert (v1 <= v2) is want


def test_version_le_invalid_type():
    v = Version(1, 2, 4)

    with pytest.raises(TypeError):
        v <= "a string"


@pytest.mark.parametrize(
    "v1, v2, want",
    [
        (Version(0, 7, 6), Version(1, 7, 6), False),
        (Version(1, 6, 6), Version(1, 7, 6), False),
        (Version(1, 7, 5), Version(1, 7, 6), False),
        (Version(1, 7, 6, "pre"), Version(1, 7, 6), False),
        (Version(1, 7, 6, "pre", "build"), Version(1, 7, 6), False),
        (Version(1, 7, 6, "pre1"), Version(1, 7, 6, "pre2"), False),
        (Version(1, 7, 6, "pre1", "build1"), Version(1, 7, 6, "pre1", "build2"), False),
        (Version(1, 7, 6, None, "build1"), Version(1, 7, 6, None, "build2"), False),
        (Version(1, 7, 6), Version(0, 7, 6), True),
        (Version(1, 7, 6), Version(1, 6, 6), True),
        (Version(1, 7, 6), Version(1, 7, 5), True),
        (Version(1, 7, 6), Version(1, 7, 6, "pre"), True),
        (Version(1, 7, 6), Version(1, 7, 6, "pre", "build"), True),
        (Version(1, 7, 6, "pre2"), Version(1, 7, 6, "pre1"), True),
        (Version(1, 7, 6, "pre2", "build2"), Version(1, 7, 6, "pre1", "build1"), True),
        (Version(1, 7, 6, None, "build2"), Version(1, 7, 6, None, "build1"), True),
    ],
)
def test_version_gt(v1: Version, v2: Version, want: bool):
    assert (v1 > v2) is want


def test_version_gt_invalid_type():
    v = Version(1, 2, 4)

    with pytest.raises(TypeError):
        v > "a string"


@pytest.mark.parametrize(
    "v1, v2, want",
    [
        (Version(0, 7, 6), Version(1, 7, 6), False),
        (Version(1, 6, 6), Version(1, 7, 6), False),
        (Version(1, 7, 5), Version(1, 7, 6), False),
        (Version(1, 7, 6, "pre"), Version(1, 7, 6), False),
        (Version(1, 7, 6, "pre", "build"), Version(1, 7, 6), False),
        (Version(1, 7, 6, "pre1"), Version(1, 7, 6, "pre2"), False),
        (Version(1, 7, 6, "pre1", "build1"), Version(1, 7, 6, "pre1", "build2"), False),
        (Version(1, 7, 6, None, "build1"), Version(1, 7, 6, None, "build2"), False),
        (Version(1, 7, 6), Version(0, 7, 6), True),
        (Version(1, 7, 6), Version(1, 7, 6), True),
        (Version(1, 7, 6), Version(1, 6, 6), True),
        (Version(1, 7, 6), Version(1, 7, 5), True),
        (Version(1, 7, 6), Version(1, 7, 6, "pre"), True),
        (Version(1, 7, 6), Version(1, 7, 6, "pre", "build"), True),
        (Version(1, 7, 6, "pre2"), Version(1, 7, 6, "pre1"), True),
        (Version(1, 7, 6, "pre2", "build2"), Version(1, 7, 6, "pre1", "build1"), True),
        (Version(1, 7, 6, None, "build2"), Version(1, 7, 6, None, "build1"), True),
    ],
)
def test_version_ge(v1: Version, v2: Version, want: bool):
    assert (v1 >= v2) is want


def test_version_ge_invalid_type():
    v = Version(1, 2, 4)

    with pytest.raises(TypeError):
        v >= "a string"


@pytest.mark.parametrize(
    "v1, v2, want",
    [
        (Version(1, 0, 0, "equal"), Version(1, 0, 0, "equal"), 0),
        (Version(1, 0, 0, "pre1"), Version(1, 0, 0, "hello1"), 0),
        (Version(1, 0, 0, "pre"), Version(1, 0, 0), -1),
        (Version(1, 0, 0), Version(1, 0, 0, "pre"), 1),
        (Version(1, 0, 0, "pre1"), Version(1, 0, 0, "pre2"), -1),
        (Version(1, 0, 0, "pre2"), Version(1, 0, 0, "pre1"), 1),
    ],
)
def test_compare_prerelease(v1: Version, v2: Version, want: int):
    assert v1._compare_prerelease(v2) == want


def test_compare_prerelease_raises_if_no_digits():
    v1 = Version(1, 0, 0, "pre")
    v2 = Version(1, 0, 0, "prepre")

    with pytest.raises(ValueError):
        v1._compare_prerelease(v2)


def test_compare_prerelease_raises_if_multiple_digits():
    v1 = Version(1, 0, 0, "pre3")
    v2 = Version(1, 0, 0, "pre3version4")

    with pytest.raises(ValueError):
        v1._compare_prerelease(v2)


def test_compare_prerelease_raises_at_the_end():
    v1 = Version(1, 0, 0, "cantcompareme")
    v2 = Version(1, 0, 0, "orme")

    with pytest.raises(ValueError):
        v1._compare_prerelease(v2)


@pytest.mark.parametrize(
    "v1, v2, want",
    [
        (Version(1, 0, 0, None, "equal"), Version(1, 0, 0, None, "equal"), 0),
        (Version(1, 0, 0, None, "build1"), Version(1, 0, 0, None, "hello1"), 0),
        (Version(1, 0, 0, None, "build"), Version(1, 0, 0), -1),
        (Version(1, 0, 0), Version(1, 0, 0, None, "build"), 1),
        (Version(1, 0, 0, None, "build1"), Version(1, 0, 0, None, "build2"), -1),
        (Version(1, 0, 0, None, "build2"), Version(1, 0, 0, None, "build1"), 1),
    ],
)
def test_compare_build(v1: Version, v2: Version, want: int):
    assert v1._compare_build(v2) == want


def test_compare_build_raises_if_no_digits():
    v1 = Version(1, 0, 0, None, "build")
    v2 = Version(1, 0, 0, None, "buildybuild")

    with pytest.raises(ValueError):
        v1._compare_build(v2)


def test_compare_build_raises_if_multiple_digits():
    v1 = Version(1, 0, 0, None, "build3")
    v2 = Version(1, 0, 0, None, "build1ver4")

    with pytest.raises(ValueError):
        v1._compare_build(v2)


def test_compare_build_raises_at_the_end():
    v1 = Version(1, 0, 0, None, "cantcompareme")
    v2 = Version(1, 0, 0, None, "orme")

    with pytest.raises(ValueError):
        v1._compare_build(v2)


@pytest.mark.parametrize(
    "v, want",
    [
        (Version(1, 2, 4), Version(2, 0, 0)),
        (Version(0, 0, 1), Version(1, 0, 0)),
        (Version(0, 0, 0), Version(1, 0, 0)),
        (Version(0, 7, 6, "pre"), Version(1, 0, 0)),
        (Version(0, 7, 6, "pre", "build"), Version(1, 0, 0)),
        (Version(999999, 999, 9999, "holy_crap"), Version(1000000, 0, 0)),
    ],
)
def test_bump_major(v: Version, want: Version):
    assert v.bump_major() == want


@pytest.mark.parametrize(
    "v, want",
    [
        (Version(1, 2, 4), Version(1, 3, 0)),
        (Version(0, 0, 1), Version(0, 1, 0)),
        (Version(0, 0, 0), Version(0, 1, 0)),
        (Version(0, 7, 6, "pre"), Version(0, 8, 0)),
        (Version(0, 7, 6, "pre", "build"), Version(0, 8, 0)),
        (Version(999999, 999, 9999, "holy_crap"), Version(999999, 1000, 0)),
    ],
)
def test_bump_minor(v: Version, want: Version):
    assert v.bump_minor() == want


@pytest.mark.parametrize(
    "v, want",
    [
        (Version(1, 2, 4), Version(1, 2, 5)),
        (Version(0, 0, 1), Version(0, 0, 2)),
        (Version(0, 0, 0), Version(0, 0, 1)),
        (Version(0, 7, 6, "pre"), Version(0, 7, 7)),
        (Version(0, 7, 6, "pre", "build"), Version(0, 7, 7)),
        (Version(999999, 999, 9999, "holy_crap"), Version(999999, 999, 10000)),
    ],
)
def test_bump_patch(v: Version, want: Version):
    assert v.bump_patch() == want


@pytest.mark.parametrize(
    "version, want",
    [
        (Version(1, 2, 4, None, None), "v1.2.4"),
        (Version(1, 2, 4), "v1.2.4"),
        (Version(2, 6, 8, "rc.2", None), "v2.6.8-rc.2"),
        (Version(2, 6, 8, "rc.2"), "v2.6.8-rc.2"),
        (Version(7, 6, 2, None, "build.123"), "v7.6.2+build.123"),
        (Version(7, 2, 1, "rc.1", "build.123"), "v7.2.1-rc.1+build.123"),
    ],
)
def test_version_to_string(version: Version, want: str):
    assert version.to_string() == want


@pytest.mark.parametrize(
    "version, want",
    [
        (Version(1, 2, 4), (1, 2, 4, None, None)),
        (Version(1, 2, 4, "rc.1"), (1, 2, 4, "rc.1", None)),
        (Version(1, 2, 4, "rc.3", "build.8"), (1, 2, 4, "rc.3", "build.8")),
    ],
)
def test_version_to_tuple(version: Version, want: VersionTuple):
    assert version.to_tuple() == want


@pytest.mark.parametrize(
    "version, want",
    [
        (
            Version(1, 2, 4),
            {
                "major": 1,
                "minor": 2,
                "patch": 4,
                "prerelease": None,
                "buildmetadata": None,
            },
        ),
        (
            Version(1, 2, 4, "rc.1"),
            {
                "major": 1,
                "minor": 2,
                "patch": 4,
                "prerelease": "rc.1",
                "buildmetadata": None,
            },
        ),
        (
            Version(1, 2, 4, "rc.1", "build.123"),
            {
                "major": 1,
                "minor": 2,
                "patch": 4,
                "prerelease": "rc.1",
                "buildmetadata": "build.123",
            },
        ),
    ],
)
def test_version_to_dict(version: Version, want: VersionDict):
    assert version.to_dict() == want


@pytest.mark.parametrize(
    "version, want",
    [
        (
            Version(1, 2, 4),
            '{"major": 1, "minor": 2, "patch": 4, "prerelease": null, "buildmetadata":'
            " null}",
        ),
        (
            Version(1, 2, 4, "rc.1", "build.2"),
            '{"major": 1, "minor": 2, "patch": 4, "prerelease": "rc.1",'
            ' "buildmetadata": "build.2"}',
        ),
    ],
)
def test_version_to_json(version: Version, want: str):
    assert version.to_json() == want


@pytest.mark.parametrize(
    "version, want",
    [
        (Version(1, 2, 4), True),
        (Version(1, 2, 3, "pre"), True),
        (Version(1, 2, 4, "pre", "build"), True),
        (Version(8, 2, 3, "rc.1", "build.123"), True),
        (Version(1, 0, 0, "alpha-a.b-c-somethinglong", "build.1-aef.1-its-okay"), True),
        (Version(1, 2, 3, "ajbas---28", "lnq==2987"), False),
        (Version(1, 2, 4, "blah198y_+-2-", "build---19790"), False),
    ],
)
def test_version_is_valid(version: Version, want: bool):
    assert version.is_valid() is want


@pytest.mark.parametrize(
    "d, want",
    [
        (
            {
                "major": 7,
                "minor": 2,
                "patch": 6,
            },
            Version(7, 2, 6),
        ),
        (
            {
                "major": 1,
                "minor": 2,
                "patch": 3,
                "prerelease": "rc.1",
            },
            Version(1, 2, 3, "rc.1"),
        ),
        (
            {
                "major": 1,
                "minor": 2,
                "patch": 3,
                "prerelease": "rc.1",
                "buildmetadata": "build.123",
            },
            Version(1, 2, 3, "rc.1", "build.123"),
        ),
    ],
)
def test_version_from_dict(d: VersionDict, want: Version):
    assert Version.from_dict(d) == want


@pytest.mark.parametrize(
    "string, want",
    [
        ("v1.2.3", Version(1, 2, 3)),
        ("1.2.3", Version(1, 2, 3)),
        ("v1.2.3-rc.1", Version(1, 2, 3, "rc.1")),
        ("v1.2.3-rc.1+build.123", Version(1, 2, 3, "rc.1", "build.123")),
    ],
)
def test_version_from_string(string: str, want: Version):
    assert Version.from_string(string) == want


def test_version_from_string_raises_on_bad_string():
    bad = "i'm not a version"

    with pytest.raises(ValueError):
        Version.from_string(bad)


@pytest.mark.parametrize("string", ["v1.2.4", "v1.2.4-rc.1", "v1.2.4-rc.1+build.123"])
def test_version_from_string_to_string_round_trip(string: str):
    assert Version.from_string(string).to_string() == string


@pytest.mark.parametrize(
    "tup, want",
    [
        ((1, 2, 4), Version(1, 2, 4)),
        ((1, 2, 4, "rc.1"), Version(1, 2, 4, "rc.1")),
        ((1, 2, 4, "rc.1", "build.123"), Version(1, 2, 4, "rc.1", "build.123")),
    ],
)
def test_version_from_tuple(tup: VersionTuple, want: Version):
    assert Version.from_tuple(tup) == want


@pytest.mark.parametrize(
    "json_string, want",
    [
        ('{"major": 1, "minor": 2, "patch": 4}', Version(1, 2, 4)),
        (
            '{"major": 1, "minor": 2, "patch": 4, "prerelease": "rc.1"}',
            Version(1, 2, 4, "rc.1"),
        ),
        (
            '{"major": 1, "minor": 2, "patch": 4, "prerelease": "rc.1","buildmetadata":'
            ' "build.2"}',
            Version(1, 2, 4, "rc.1", "build.2"),
        ),
    ],
)
def test_version_from_json(json_string: str, want: Version):
    assert Version.from_json(json_string) == want
